#include "DXInput.h"

void DXInput::DI_Term()
{ 
	if (m_lpDI) 
	{ 
		if (m_lpDIDeviceKB)
		{ 
			// Always unacquire device before calling Release().
			m_lpDIDeviceKB->Unacquire(); 
			m_lpDIDeviceKB->Release();
			m_lpDIDeviceKB = NULL; 
		} 
		m_lpDI->Release();
		m_lpDI = NULL; 
	} 
}

bool DXInput::DI_Init( HINSTANCE hInst, HWND hWnd ) 
{ 
	HRESULT hr; 

	// Create the DirectInput object. 
	hr = DirectInputCreateEx(hInst, DIRECTINPUT_VERSION, IID_IDirectInput7, (void**)&m_lpDI, NULL); 

	if FAILED(hr)
	{
		Log( __FILE__, __LINE__, "Failed DXInput::DI_Init() call - DirectInputCreateEx" );
		return FALSE; 
	}

	// Retrieve a pointer to an IDirectInputDevice7 interface for the Keyboard
	hr = m_lpDI->CreateDeviceEx(GUID_SysKeyboard, IID_IDirectInputDevice7, (void**)&m_lpDIDeviceKB, NULL); 
	if FAILED(hr) 
	{
		Log( __FILE__, __LINE__, "Failed DXInput::DI_Init() call - CreateDeviceEx (Keyboard)" );
		DI_Term();
		return FALSE;
	}

	// Now that you have an IDirectInputDevice7 interface, get 
	// it ready to use. 

	// Set the data format using the predefined keyboard data 
	// format provided by the DirectInput object for keyboards. 
	hr = m_lpDIDeviceKB->SetDataFormat(&c_dfDIKeyboard); 
	
	if FAILED(hr) 
	{ 
		Log( __FILE__, __LINE__, "Failed DXInput::DI_Init() call - SetDataFormat (Keyboard)" );
		DI_Term(); 
		return FALSE; 
	} 

	// Set the cooperative level 
	hr = m_lpDIDeviceKB->SetCooperativeLevel(hWnd, DISCL_BACKGROUND | DISCL_NONEXCLUSIVE); 
	if FAILED(hr) 
	{ 
		Log( __FILE__, __LINE__, "Failed DXInput::DI_Init() call - SetCooperativeLevel (Keyboard)" );
		DI_Term(); 
		return FALSE; 
	} 

	// Get access to the input device. 
	hr = m_lpDIDeviceKB->Acquire(); 

	if FAILED(hr) 
	{ 
		Log( __FILE__, __LINE__, "Failed DXInput::DI_Init() call - Acquire (Keyboard)" );
		DI_Term(); 
		return FALSE; 
	} 

	return TRUE; 
}

void DXInput::ProcessKBInput() 
{ 
    HRESULT  hr; 
 
    hr = m_lpDIDeviceKB->GetDeviceState(sizeof(buffer),(LPVOID)&buffer); 
    if FAILED(hr) 
    { 
         // If it failed, the device has probably been lost. 
         // We should check for (hr == DIERR_INPUTLOST) 
         // and attempt to reacquire it here. 
         return; 
    } 
}

int DXInput::KeyPressed( int Key )
{
	return KEYDOWN( buffer, Key );
}
