#include "DXMain.h"
#include <stdio.h>

DXMain::DXMain()
{
	lpDD = NULL;
	Back = NULL;
	Prim = NULL;
}

DXMain::~DXMain()
{
	Surfs.RemoveAllSurfaces();
	SAFEDXREL( Back );
	SAFEDXREL( Prim );
	SAFEDXREL( lpDD );
}

bool DXMain::InitDirectDraw( HWND hWnd, int width, int height, int bpp, bool windowed )
{
	DDSURFACEDESC2 ddsd;
	DDSCAPS2 ddsCaps;
	HRESULT ddrval;

	Log( __FILE__, __LINE__, "InitDirectDraw Entered" );

	ddrval = DirectDrawCreateEx(NULL, ( void ** ) &lpDD, IID_IDirectDraw7, NULL);

	m_Windowed = windowed;
	if( ddrval != DD_OK )
	{
		Log( __FILE__, __LINE__, "\tFailed to create DirectX 7 Interface" );
		return false;
	}

	if( !windowed ) // if full screen mode required
	{
		// Set the windows display properies

		long lStyle, lStyleNew, lSt;

		lStyle = GetWindowLong( hWnd, GWL_STYLE );

		if( lStyle != 0 )
		{
			lStyleNew = WS_OVERLAPPED | WS_VISIBLE;
			lSt = SetWindowLong( hWnd, GWL_STYLE, lStyleNew );
			if( lSt != lStyle )
				Log( __FILE__, __LINE__, "SetWindowLong() error??" );
		}
		else
			Log( __FILE__, __LINE__, "GetWindowLong() error ??" );

		// Now on to the good stuff - Initalise DirectX
		ddrval = lpDD->SetCooperativeLevel( hWnd, DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE );

		if( ddrval != DD_OK )
		{
			Log( __FILE__, __LINE__, "\tFailed to set Coop level");
			SAFEDXREL( lpDD );
			return false;
		}

		ddrval = lpDD->SetDisplayMode( width, height, bpp, 0, 0 );
		if( ddrval != DD_OK )
		{
			char t[256];

			sprintf( t, "Failed to set mode %dx%d %dbpp, attempting 640x480 16bpp", width, height, bpp );
			Log( __FILE__, __LINE__, t );
			ddrval = lpDD->SetDisplayMode( 640, 480, 16, 0, 0 );
			if( ddrval != DD_OK )
			{
				Log( __FILE__, __LINE__, "Failed to set default display mode, exiting" );
				SAFEDXREL( lpDD );
				return false;
			}
			Log( __FILE__, __LINE__, "Set mode to 640x480 16bpp" );
		}

		DXSTRUCTINIT( ddsd );

		ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX; 
		ddsd.dwBackBufferCount = 1;

		ddrval = lpDD->CreateSurface( &ddsd, &Prim, NULL );
		if( ddrval != DD_OK )
		{
			Log( __FILE__, __LINE__, "Failed to create primary surface, exiting" );
			Log( __FILE__, __LINE__, DDErrorString( ddrval ) );
			SAFEDXREL( lpDD );
			return false;
		}

		ZeroMemory( &ddsCaps, sizeof( ddsCaps ) );
		ddsCaps.dwCaps = DDSCAPS_BACKBUFFER; 

		ddrval = Prim->GetAttachedSurface( &ddsCaps, &Back);
		if( ddrval != DD_OK )
		{
			Log( __FILE__, __LINE__, "Failed to create backbuffer, exiting" );
			Log( __FILE__, __LINE__, DDErrorString( ddrval ) );
			SAFEDXREL( Prim );
			SAFEDXREL( lpDD );
			return false;
		}

		m_ScreenWidth = width;
		m_ScreenHeight = height;
	}
	else
	{
		RECT rc = { 0, 0, width, height };
		RECT rcWork;

		Log( __FILE__, __LINE__, "WARNING - Windowed mode code is in BETA state!!." );

		// Set the windows display properies

		long lStyle, lStyleNew, lSt;

		lStyle = GetWindowLong( hWnd, GWL_STYLE );

		if( lStyle != 0 )
		{
			lStyleNew = WS_OVERLAPPED | WS_CAPTION | WS_THICKFRAME | WS_SYSMENU | WS_VISIBLE | WS_MAXIMIZEBOX;
			lSt = SetWindowLong( hWnd, GWL_STYLE, lStyleNew );


			AdjustWindowRectEx( &rc, GetWindowStyle(hWnd), GetMenu(hWnd) != NULL, GetWindowExStyle(hWnd) );

			SetWindowPos( hWnd, NULL, 0, 0, rc.right - rc.left, rc.bottom - rc.top, SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE );

			SetWindowPos( hWnd, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE | SWP_NOACTIVATE );

			SystemParametersInfo( SPI_GETWORKAREA, 0, &rcWork, 0 );
			GetWindowRect( hWnd, &rc );
			if( rc.left < rcWork.left ) rc.left = rcWork.left;
			if( rc.top  < rcWork.top )  rc.top  = rcWork.top;
			SetWindowPos( hWnd, NULL, rc.left, rc.top, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE );

			if( lSt != lStyle )
				Log( __FILE__, __LINE__, "SetWindowLong() error??" );
		}
		else
			Log( __FILE__, __LINE__, "GetWindowLong() error ??" );

		// Now on to the good stuff - Initalise DirectX
		ddrval = lpDD->SetCooperativeLevel( hWnd, DDSCL_NORMAL );

		if( ddrval != DD_OK )
		{
			Log(__FILE__, __LINE__, "\tFailed to set Coop level");
			SAFEDXREL( lpDD );
			return false;
		}

		m_hMainWnd = hWnd;

		DXSTRUCTINIT( ddsd );

		lpDD->GetDisplayMode(&ddsd);
		if(ddsd.ddpfPixelFormat.dwRGBBitCount < (DWORD)bpp)
		{
			MessageBox(hWnd,"This demo was designed for 16 bit mode","IsoEngine Error",MB_OK);
			return false;
		}
		else
			bpp = ddsd.ddpfPixelFormat.dwRGBBitCount;

		RECT rcClient;
		GetClientRect( m_hMainWnd, &rcClient );

		m_dwClientWidth = rcClient.right;
		m_dwClientHeight = rcClient.bottom;

		DXSTRUCTINIT( ddsd );

		ddsd.dwFlags = DDSD_CAPS;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE; 

		ddrval = lpDD->CreateSurface( &ddsd, &Prim, NULL );
		if( ddrval != DD_OK )
		{
			Log( __FILE__, __LINE__, "Failed to create primary surface, exiting" );
			Log( __FILE__, __LINE__, DDErrorString( ddrval ) );
			SAFEDXREL( lpDD );
			return false;
		}

//		ddrval = lpDD->CreateClipper( 0, &lpClip, NULL );
//		if( ddrval != DD_OK )
//		{
//			Log( __FILE__, __LINE__, "Failed to create clipper object, exiting" );
//			Log( __FILE__, __LINE__, DDErrorString( ddrval ) );
//			SAFEDXREL( Prim );
//			SAFEDXREL( lpDD );
//			return false;
//		}
//
//		ddrval = lpClip->SetHWnd( 0, hWnd );
//		if( ddrval != DD_OK )
//		{
//			Log( __FILE__, __LINE__, "Failed to set clipper object's co-ords, exiting" );
//			Log( __FILE__, __LINE__, DDErrorString( ddrval ) );
//			SAFEDXREL( lpClip );
//			SAFEDXREL( Prim );
//			SAFEDXREL( lpDD );
//			return false;
//		}
//
//		ddrval = Prim->SetClipper( lpClip );
//		if( ddrval != DD_OK )
//		{
//			Log( __FILE__, __LINE__, "Failed to attach clipper object, exiting" );
//			Log( __FILE__, __LINE__, DDErrorString( ddrval ) );
//			SAFEDXREL( lpClip );
//			SAFEDXREL( Prim );
//			SAFEDXREL( lpDD );
//			return false;
//		}

		DXSTRUCTINIT( ddsd );

		ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
		ddsd.dwHeight = height;
		ddsd.dwWidth = width;

		ddrval = lpDD->CreateSurface( &ddsd, &Back, NULL );

		if(ddrval != DD_OK)
		{
			Log( __FILE__, __LINE__, "Failed to create backbuffer, exiting" );
			Log( __FILE__, __LINE__, DDErrorString( ddrval ) );
			SAFEDXREL( Prim );
			SAFEDXREL( lpDD );
			return NULL;
		}

		m_ScreenWidth = width;
		m_ScreenHeight = height;
	}

	Surfs.SetDD( lpDD );
	Surfs.SetBPP( bpp );
	m_bpp = bpp;

	Log( __FILE__, __LINE__, "DirectX 7 Init OK" );

	return true;
}

bool DXMain::DeInitDirectDraw()
{
	Surfs.RemoveAllSurfaces();
	SAFEDXREL( Back );
	SAFEDXREL( Prim );
	SAFEDXREL( lpDD );
	return true;
}

void DXMain::Flip()
{
	HRESULT ddrval;

	if( Prim )
	{
		if( !m_Windowed )
		{
			ddrval = Prim->Flip( NULL, 0 );
			if( ddrval != DD_OK )
			{
				if( ddrval == DDERR_SURFACELOST )
				{
					Log( __FILE__, __LINE__, "Flip - Lost Surfaces" );
					Prim->Restore();
				}
				else
				{
					char t[256];

					sprintf( t, "Flip Error - %s", DDErrorString( ddrval ) );

					Log( __FILE__, __LINE__, t );
				}
			}
		}
		else
		{
			POINT pt = { 0, 0 };
			RECT rcD = { 0, 0, m_dwClientWidth, m_dwClientHeight };
			RECT rcS = { 0, 0, m_ScreenWidth, m_ScreenHeight };
			RECT rcClient;

			GetClientRect( m_hMainWnd, &rcClient );

			m_dwClientWidth = rcClient.right;
			m_dwClientHeight = rcClient.bottom;
			rcD.right = m_dwClientWidth;
			rcD.bottom = m_dwClientHeight;

			ClientToScreen( m_hMainWnd, &pt );

			OffsetRect( &rcD, pt.x, pt.y );

			ddrval = Prim->Blt( &rcD, Back, &rcS, DDBLT_WAIT, NULL );

			if( ddrval != DD_OK )
				Log( __FILE__, __LINE__, DDErrorString( ddrval ) );
		}
	}
}

pSurface DXMain::CreateSurface(int width, int height, int mem, bool Alpha)
{
	return Surfs.CreateSurface( width, height, mem, Alpha );
}

pSurface DXMain::LoadBitmap(char *Filename, int Mem)
{
	return Surfs.LoadBitmap( Filename, Mem );
}

pSurface DXMain::LoadPCX(char *Filename, int Mem)
{
	return Surfs.LoadPCX( Filename, Mem );
}

pSurface DXMain::LoadTGA(char *Filename, int Mem)
{
	return Surfs.LoadTGA( Filename, Mem );
}

void DXMain::Blit2Screen(pSurface Src, RECT SrcRct, int x, int y )
{
	RECT DRct;

	DRct.top    = y;
	DRct.left   = x;
	DRct.right  = x + ( SrcRct.right - SrcRct.left );
	DRct.bottom = y + ( SrcRct.bottom - SrcRct.top );

	if( DRct.top < 0 )
	{
		SrcRct.top += -DRct.top;
		DRct.top = 0;
	}
	if( DRct.left < 0 )
	{
		SrcRct.left += -DRct.left;
		DRct.left = 0;
	}
	if( DRct.right > m_ScreenWidth )
	{
		SrcRct.right += ( m_ScreenWidth - DRct.right );
		DRct.right = m_ScreenWidth;
	}
	if( DRct.bottom > m_ScreenHeight )
	{
		SrcRct.bottom += ( m_ScreenHeight - DRct.bottom );
		DRct.bottom = m_ScreenHeight;
	}

	if( ( DRct.top < m_ScreenHeight ) && ( DRct.bottom > 0 ) && ( DRct.left < m_ScreenWidth ) && ( DRct.right > 0 ) )
		Blit( Src, SrcRct, Back, DRct );
}

void DXMain::Blit2Screen(pSurface Src, int x, int y )
{
	RECT SrcRct = { 0, 0, Src->Width, Src->Height };

	Blit2Screen( Src, SrcRct, x, y );
}

void DXMain::Blit2Screen2(pSurface Src, pSurface Backup, RECT SrcRct, int x, int y, USHORT col )
{
	RECT DRct;

	DRct.top    = y;
	DRct.left   = x;
	DRct.right  = x + ( SrcRct.right - SrcRct.left );
	DRct.bottom = y + ( SrcRct.bottom - SrcRct.top );

	if( DRct.top < 0 )
	{
		SrcRct.top += -DRct.top;
		DRct.top = 0;
	}
	if( DRct.left < 0 )
	{
		SrcRct.left += -DRct.left;
		DRct.left = 0;
	}
	if( DRct.right > m_ScreenWidth )
	{
		SrcRct.right += ( m_ScreenWidth - DRct.right );
		DRct.right = m_ScreenWidth;
	}
	if( DRct.bottom > m_ScreenHeight )
	{
		SrcRct.bottom += ( m_ScreenHeight - DRct.bottom );
		DRct.bottom = m_ScreenHeight;
	}

	if( ( DRct.top < m_ScreenHeight ) && ( DRct.bottom > 0 ) && ( DRct.left < m_ScreenWidth ) && ( DRct.right > 0 ) )
		SpecialBlit( Src, Backup, SrcRct, DRct, col );
}

void DXMain::Blit(pSurface Src, RECT SrcRct, pSurface Dest, RECT DestRct)
{
	if( ( Dest ) && ( Src ) )
		if( Src->UsesColorKey )
			Dest->lpDDS->Blt( &DestRct, Src->lpDDS, &SrcRct, DDBLT_WAIT | DDBLT_KEYSRC, NULL );
		else
			Dest->lpDDS->Blt( &DestRct, Src->lpDDS, &SrcRct, DDBLT_WAIT, NULL );
	else
	{
		if( ( !Dest ) && (  Src ) )
			Log1( __FILE__, __LINE__, "DXMain::Blit() - Tried to blit to non-existing surface", Dest->Name );
		if( (  Dest ) && ( !Src ) )
			Log1( __FILE__, __LINE__, "DXMain::Blit() - Tried to blit from non-existing surface", Src->Name );
		if( ( !Dest ) && ( !Src ) )
			Log2( __FILE__, __LINE__, "DXMain::Blit() - Tried to blit from non-existing surface to non-existing surface", Dest->Name, Src->Name );
	}
}

void DXMain::Blit(pSurface Src, RECT SrcRct, LPDIRECTDRAWSURFACE7 Dest, RECT DestRct)
{
	HRESULT ddrval;

	if( Dest )
		if( Src->UsesColorKey )
		{
			ddrval = Dest->Blt( &DestRct, Src->lpDDS, &SrcRct, DDBLT_WAIT | DDBLT_KEYSRC, NULL );
			if( ddrval != DD_OK )
			{
				Log( __FILE__, __LINE__, "Blit error" );
				Log( __FILE__, __LINE__, DDErrorString( ddrval ) );
			}
		}
		else
		{
			ddrval = Dest->Blt( &DestRct, Src->lpDDS, &SrcRct, DDBLT_WAIT, NULL );
			if( ddrval != DD_OK )
			{
				Log( __FILE__, __LINE__, "Blit error" );
				Log( __FILE__, __LINE__, DDErrorString( ddrval ) );
			}
		}
	else
		Log( __FILE__, __LINE__, "DXMain::Blit() - Tried to blt to none existing surface!!" );
}

bool DXMain::SetColorKey(pSurface Surf, DWORD Low, DWORD High)
{
	DDCOLORKEY ddc;
	HRESULT ddrval;

	ddc.dwColorSpaceHighValue = High;
	ddc.dwColorSpaceLowValue  = Low;

	if( ( ddrval = Surf->lpDDS->SetColorKey( DDCKEY_SRCBLT , &ddc ) ) == DD_OK )
	{
		Surf->UsesColorKey = true;
		return true;
	}
	else
	{
		char t[256];

		sprintf( t, "DXMain::SetColorKey() - Failed to set Colour key for surface %s. Reason : %s", Surf->Name, DDErrorString( ddrval ) );
		Log( __FILE__, __LINE__, t );

		return false;
	}
}

void DXMain::ClrScreen()
{
	DDBLTFX ddfx;

	DXSTRUCTINIT( ddfx );
	ddfx.dwFillColor = 0;

	Back->Blt( NULL, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddfx );
}

void DXMain::SpecialBlit( pSurface Source, pSurface Backup, RECT SrcRect, RECT DestRect, USHORT col )
{
	int xheight;
	int xwidth;
	DDSURFACEDESC2 ddsdSource;
	DDSURFACEDESC2 ddsdDest;
	DDSURFACEDESC2 ddsdBack;
	int ipitchdest, ipitchsrc, ipitchback;
	HRESULT ddrval;
	char t[256];

	DXSTRUCTINIT( ddsdSource );
	DXSTRUCTINIT( ddsdDest );
	DXSTRUCTINIT( ddsdBack );

	xwidth = SrcRect.right - SrcRect.left;
	xheight = SrcRect.bottom - SrcRect.top;

	ddrval = Source->lpDDS->Lock( &SrcRect, &ddsdSource, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL );
	if( ddrval != DD_OK )
	{
		sprintf( t, "DXMain::SpecialBlit() - Failed to lock Source : %s", DDErrorString( ddrval ) );
		Log( __FILE__, __LINE__, t );
	}

	ddrval = Back->Lock( &DestRect, &ddsdDest, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL );
	if( ddrval != DD_OK )
	{
		sprintf( t, "DXMain::SpecialBlit() - Failed to lock Dest : %s", DDErrorString( ddrval ) );
		Log( __FILE__, __LINE__, t );
	}

	ddrval = Backup->lpDDS->Lock( &DestRect, &ddsdBack, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL );
	if( ddrval != DD_OK )
	{
		sprintf( t, "DXMain::SpecialBlit() - Failed to lock Backup : %s", DDErrorString( ddrval ) );
		Log( __FILE__, __LINE__, t );
	}

	Source->lpDDS->Unlock( NULL );
	Back->Unlock( NULL );
	Backup->lpDDS->Unlock( NULL );

	ipitchsrc = ddsdSource.lPitch - ( xwidth << 1 );
	ipitchdest = ddsdDest.lPitch - ( xwidth << 1 );
	ipitchback = ddsdBack.lPitch - ( xwidth << 1 );

	_asm
	{
		xor eax, eax          // Initialize the variables. This is VC++ 6
		xor ebx, ebx          // inline Assembler. Mostly this is designed
		mov ecx, xheight      // for a DirectDraw locked surface.
		xor edx, edx
		mov esi, ddsdSource.lpSurface   // This variable should be the pointer to the source surface.
		mov edi, ddsdDest.lpSurface     // This variable should be the pointer to the dest surface.
		mov ebx, edi;
		mov edi, ddsdBack.lpSurface     // This variable should be the pointer to the backup surface.

		mov dx, col

		heightLoop16bit565:
		push ecx
		mov ecx, xwidth
		widthLoop16bit565:
		mov ax, [esi]         // Reads the 16-bit pixel (this is designed for 565 RGB).
		test ax, ax           // Tests it with an AND to itself. Sets flags, basically.
		jz exLoop16bit565     // Skips if pure black (transparent in my game). Use cmp ax, value
							  // if black isn't transparent for you. Take out this line and the previous
							  // if you don't need a transparent additive effect.

		mov [edi], dx
		push edi
		mov edi, ebx
//		Comment the next line for debugging
		mov [edi], ax
//		Uncomment the next line for debugging
//		mov [edi], dx
		pop edi

		exLoop16bit565:
		add esi, 2            // Add 2 bytes to the source pointer, to go to the next pixel.
		add edi, 2            // Ditto with the destination pointer.
		add ebx, 2
		dec ecx               // Decrease ECX. As you may have noticed, height was PUSHed in ECX,
							  // and we did that to make room for width. x86 needs more registers.
		jnz widthLoop16bit565      // If we aren't out of pixels on that row, go back for another pixel.
		pop ecx               // Pulls the height from the stack. Height is now in ECX for a sec.
		add esi, ipitchsrc    // ipitchsrc should be the pitch DirectDraw gave you in Lock()
		add edi, ipitchdest   // minus twice the source width. ipitchdest should be the same thing for
		add ebx, ipitchback
		dec ecx               // the destination surface. We're almost done, now we are decreasing
		jnz heightLoop16bit565      // the height and looping back for the next row.
	}
}
