#include "DXSurface.h"

DXSurface::DXSurface()
{
	MLog( __FILE__, __LINE__, "Init DXSurface Class" );
}

DXSurface::~DXSurface()
{
	MLog( __FILE__, __LINE__, "DeInit DXSurface Class" );
	Log( __FILE__, __LINE__, "Freeing Surfaces" );
	RemoveAllSurfaces();
}

pSurface DXSurface::CreateSurface( int Width, int Height, int Mem, bool Alpha )
{
	DDSURFACEDESC2 ddsd;
	pSurface temp;
	HRESULT ddrval;

	temp = new Surface;

	temp->Width = Width;
	temp->Height = Height;
	temp->Bit = NULL;
	temp->Next = NULL;
	temp->Prev = NULL;
	temp->UsesColorKey = false;

	DXSTRUCTINIT( ddsd );

	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;

	if( Mem == SystemMemory )
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY; 
	else
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_VIDEOMEMORY;

	ddsd.dwHeight = Height;
	ddsd.dwWidth = Alpha ? Width << 1 : Width;

	ddrval = m_lpDD->CreateSurface( &ddsd, &temp->lpDDS, NULL );

	if(ddrval != DD_OK)
	{
		Log( __FILE__, __LINE__, "Error creating offscreen surface" );
		Log( __FILE__, __LINE__, DDErrorString( ddrval ) );
		delete temp;
		return NULL;
	}

	AddSurface( temp );

	MLog( __FILE__, __LINE__, "Alloc Surface" );

	return temp;
}

bool DXSurface::RemoveSurface( pSurface surf )
{
	if( surf )
	{
		if( surf == SurfList )
		{
			if( surf->Next )
				surf->Next->Prev = surf->Prev;
			if( surf->Prev )
				surf->Prev->Next = surf->Next;
			if( surf->Next )
				SurfList = surf->Next;
			else
				if( surf->Prev )
					SurfList = surf->Prev;
				else
					SurfList = NULL; // this should only happen if you remove the only surface in the list
		}
		else
		{
			if( surf->Next )
				surf->Next->Prev = surf->Prev;
			if( surf->Prev )
				surf->Prev->Next = surf->Next;
		}

		SAFEDXREL( surf->lpDDS );
		DeleteObject( surf->Bit );
		MLog1( __FILE__, __LINE__, "Removed Surface", surf->Name );
		delete surf;
	}

	return true;
}

bool DXSurface::RemoveAllSurfaces()
{
	pSurface a = SurfList, b;

	if( a )
	{
		while( a->Next )
			a = a->Next;

		while( a )
		{
			b = a->Prev;
			RemoveSurface( a );
			a = b;
		}
		
		SurfList = a;
	}

	return true;
}

void DXSurface::AddSurface( pSurface temp )
{
	pSurface a = SurfList;

	if( a )
	{
		while( a->Next )
			a = a->Next;

		a->Next = temp;
		temp->Prev = a;
	}
	else
		SurfList = temp;
}

void DXSurface::SetDD(LPDIRECTDRAW7 lpDD)
{
	m_lpDD = lpDD;
}

void DXSurface::SetBPP(int bpp)
{
	m_bpp = bpp;
}

pSurface DXSurface::LoadBitmap(char *Filename, int mem)
{
	pSurface temp;
	HBITMAP Bit;
	BITMAP BM;
	HDC hDCImage, hDC;
	HRESULT ddrval;
	char t[256];

	Bit = (HBITMAP) LoadImage( NULL, Filename, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE | LR_CREATEDIBSECTION );
	if( Bit == NULL )
	{
		sprintf( t, "DXSURFACE::LoadBitmap - Failed to load bitmap : %s", Filename );
		Log( __FILE__, __LINE__, t );
		return NULL;
	}

	GetObject( Bit, sizeof( BM ), &BM );

	temp = CreateSurface( BM.bmWidth, BM.bmHeight, mem );

	strcpy( temp->Name, Filename );

	MLog1( __FILE__, __LINE__, "Added surface", Filename );

	temp->Bit = Bit;

	ddrval = temp->lpDDS->GetDC( &hDC );

	if( ddrval != DD_OK )
	{
		Log( __FILE__, __LINE__, "DXSURFACE::LoadBitmap - Failed GetDC() Call" );
		DeleteObject( Bit );
		delete temp;
		return NULL;
	}

	hDCImage = CreateCompatibleDC( NULL );
	SelectObject( hDCImage, Bit );

	BitBlt( hDC, 0, 0, BM.bmWidth, BM.bmHeight, hDCImage, 0, 0, SRCCOPY );
	temp->lpDDS->ReleaseDC( hDC );

	DeleteDC( hDCImage );

	return temp;
}

pSurface DXSurface::LoadPCX(char *Filename, int mem)
{
	switch( m_bpp )
	{
	case 16:
		return LoadPCX16( Filename, mem );
		break;
	case 24:
		return LoadPCX24( Filename, mem );
		break;
	case 32:
		return LoadPCX32( Filename, mem );
		break;
	}

	return NULL;
}

pSurface DXSurface::LoadPCX16(char *Filename, int mem)
{
	FILE *f;
	PCXHeader hdr;
	char t[256];
	char fc;
	pSurface temp;
	int i,j,k,l,m;
	DDPIXELFORMAT ddPixFormat;
	ULONG rbits, gbits, bbits;
	USHORT rshift, gshift, bshift;
	DDSURFACEDESC2 ddSurf;
	HRESULT ddrval;
	USHORT bits, shift, col;

	f = fopen( Filename, "rb" );
	if( f == NULL )
	{
		Log( __FILE__, __LINE__, "DXSurface::LoadPCX16() - Could not open file" );
		return NULL;
	}

	fread( &hdr, sizeof( _PCX_HEADER ), 1, f );

	temp = CreateSurface( ( hdr.XMax - hdr.XMin ) + 1, ( hdr.YMax - hdr.YMin ) + 1, mem );
	if( temp == NULL )
	{
		Log( __FILE__, __LINE__, "DXSurface::LoadPCX16() - Could not create surface" );
		fclose( f );
		return NULL;
	}

	strcpy( temp->Name, Filename );

	MLog1( __FILE__, __LINE__, "Added surface", Filename );

	DXSTRUCTINIT( ddPixFormat );

	temp->lpDDS->GetPixelFormat( &ddPixFormat );

	DXSTRUCTINIT( ddSurf );

	ddrval = temp->lpDDS->Lock( NULL, &ddSurf, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL );
	if( ddrval != DD_OK )
	{
		sprintf( t, "DXSurface::LoadPCX16() - Failed to lock surface : %s", DDErrorString( ddrval ) );
		Log( __FILE__, __LINE__, t );
	}

	temp->lpDDS->Unlock( NULL );

	rbits = gbits = bbits = rshift = gshift = bshift = 0;

	if( ( ddPixFormat.dwGBitMask >> 5 ) == 0x3F )
	{
		rbits = 0xF8;
		gbits = 0xFC;
		bbits = 0xF8;
		rshift = 8;
		gshift = 3;
		bshift = 3;
	}
	else
	{
		rbits = 0xF8;
		gbits = 0xF8;
		bbits = 0xF8;
		rshift = 7;
		gshift = 3;
		bshift = 3;
	}

	k = 0;

	for( i = 0; i < ( ( hdr.YMax - hdr.YMin ) + 1 ); i ++ )
	{
		for( k = 0; k < hdr.NPlanes; k ++ )
		{
			m = ( i * ddSurf.lPitch );

			_asm
			{
				mov edi, ddSurf.lpSurface
				mov eax, m
				add edi, eax
			};

			switch( k )
			{
			case 0:
				bits  = (USHORT)rbits;
				shift = (USHORT)rshift;
				break;
			case 1:
				bits  = (USHORT)gbits;
				shift = (USHORT)gshift;
				break;
			case 2:
				bits  = (USHORT)bbits;
				shift = (USHORT)bshift;
				break;
			}

			for( j = 0; j < hdr.BytesPerLine; j ++ )
			{
				fc = fgetc( f );
				if( ( fc & 0xC0 ) == 0xC0)
				{
					l = (fc & 0x3F);

					fc = fgetc( f );

					if( k != 2 )
						col = ( fc & bits ) << shift;
					else
						col = ( fc & bits ) >> shift;

					_asm
					{
						mov ecx, l
						InnerLoop:
						mov ax,[edi]
						or  ax, col
						mov [edi],ax
						add edi, 2
						inc j
						dec cx
						jnz InnerLoop
						dec j
					};
				}
				else
				{
					if( k != 2 )
						col = ( fc & bits ) << shift;
					else
						col = ( fc & bits ) >> shift;

					_asm
					{
						mov ax,[edi]
						or  ax, col
						mov [edi],ax
						add edi, 2
					};
				}
			}
		}

	}

	fclose( f );

	return temp;
}

pSurface DXSurface::LoadPCX24(char *Filename, int mem)
{
	FILE *f;
	PCXHeader hdr;
	char t[256];
	char fc;
	pSurface temp;
	int i,j,k,l,m;
	DDPIXELFORMAT ddPixFormat;
	ULONG rbits, gbits, bbits;
	USHORT rshift, gshift, bshift;
	DDSURFACEDESC2 ddSurf;
	HRESULT ddrval;
	DWORD bits, shift, col;

	f = fopen( Filename, "rb" );
	if( f == NULL )
	{
		Log( __FILE__, __LINE__, "DXSurface::LoadPCX24() - Could not open file" );
		return NULL;
	}

	fread( &hdr, sizeof( _PCX_HEADER ), 1, f );

	temp = CreateSurface( ( hdr.XMax - hdr.XMin ) + 1, ( hdr.YMax - hdr.YMin ) + 1, mem );
	if( temp == NULL )
	{
		Log( __FILE__, __LINE__, "DXSurface::LoadPCX24() - Could not create surface" );
		fclose( f );
		return NULL;
	}

	strcpy( temp->Name, Filename );

	MLog1( __FILE__, __LINE__, "Added surface", Filename );

	DXSTRUCTINIT( ddPixFormat );

	temp->lpDDS->GetPixelFormat( &ddPixFormat );

	DXSTRUCTINIT( ddSurf );

	ddrval = temp->lpDDS->Lock( NULL, &ddSurf, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL );
	if( ddrval != DD_OK )
	{
		sprintf( t, "DXSurface::LoadPCX24() - Failed to lock surface : %s", DDErrorString( ddrval ) );
		Log( __FILE__, __LINE__, t );
	}

	temp->lpDDS->Unlock( NULL );

	rbits = gbits = bbits = rshift = gshift = bshift = 0;

	rbits = 0xFF;
	gbits = 0xFF;
	bbits = 0xFF;
	rshift = 16;
	gshift =  8;
	bshift =  0;

	k = 0;

	for( i = 0; i < ( ( hdr.YMax - hdr.YMin ) + 1 ); i ++ )
	{
		for( k = 0; k < hdr.NPlanes; k ++ )
		{
			m = ( i * ddSurf.lPitch );

			_asm
			{
				mov edi, ddSurf.lpSurface
				mov eax, m
				add edi, eax
			};

			switch( k )
			{
			case 0:
				bits  = (USHORT)rbits;
				shift = (USHORT)rshift;
				break;
			case 1:
				bits  = (USHORT)gbits;
				shift = (USHORT)gshift;
				break;
			case 2:
				bits  = (USHORT)bbits;
				shift = (USHORT)bshift;
				break;
			}

			for( j = 0; j < hdr.BytesPerLine; j ++ )
			{
				fc = fgetc( f );
				if( ( fc & 0xC0 ) == 0xC0)
				{
					l = (fc & 0x3F);

					fc = fgetc( f );

					col = ( fc & bits ) << shift;

					_asm
					{
						mov ecx, l
						InnerLoop:
						mov eax,[edi]
						or  eax, col
						mov [edi],eax
						add edi, 3
						inc j
						dec cx
						jnz InnerLoop
						dec j
					};
				}
				else
				{
					col = ( fc & bits ) << shift;

					_asm
					{
						mov eax,[edi]
						or  eax, col
						mov [edi],eax
						add edi, 3
					};
				}
			}
		}

	}

	fclose( f );

	return temp;
}

pSurface DXSurface::LoadPCX32(char *Filename, int mem)
{
	FILE *f;
	PCXHeader hdr;
	char t[256];
	char fc;
	pSurface temp;
	int i,j,k,l,m;
	DDPIXELFORMAT ddPixFormat;
	ULONG rbits, gbits, bbits;
	USHORT rshift, gshift, bshift;
	DDSURFACEDESC2 ddSurf;
	HRESULT ddrval;
	DWORD bits, shift, col;

	f = fopen( Filename, "rb" );
	if( f == NULL )
	{
		Log( __FILE__, __LINE__, "DXSurface::LoadPCX24() - Could not open file" );
		return NULL;
	}

	fread( &hdr, sizeof( _PCX_HEADER ), 1, f );

	temp = CreateSurface( ( hdr.XMax - hdr.XMin ) + 1, ( hdr.YMax - hdr.YMin ) + 1, mem );
	if( temp == NULL )
	{
		Log( __FILE__, __LINE__, "DXSurface::LoadPCX24() - Could not create surface" );
		fclose( f );
		return NULL;
	}

	strcpy( temp->Name, Filename );

	MLog1( __FILE__, __LINE__, "Added surface", Filename );

	DXSTRUCTINIT( ddPixFormat );

	temp->lpDDS->GetPixelFormat( &ddPixFormat );

	DXSTRUCTINIT( ddSurf );

	ddrval = temp->lpDDS->Lock( NULL, &ddSurf, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL );
	if( ddrval != DD_OK )
	{
		sprintf( t, "DXSurface::LoadPCX24() - Failed to lock surface : %s", DDErrorString( ddrval ) );
		Log( __FILE__, __LINE__, t );
	}

	temp->lpDDS->Unlock( NULL );

	rbits = gbits = bbits = rshift = gshift = bshift = 0;

	rbits = 0xFF;
	gbits = 0xFF;
	bbits = 0xFF;
	rshift = 16;
	gshift =  8;
	bshift =  0;

	k = 0;

	for( i = 0; i < ( ( hdr.YMax - hdr.YMin ) + 1 ); i ++ )
	{
		for( k = 0; k < hdr.NPlanes; k ++ )
		{
			m = ( i * ddSurf.lPitch );

			_asm
			{
				mov edi, ddSurf.lpSurface
				mov eax, m
				add edi, eax
			};

			switch( k )
			{
			case 0:
				bits  = (USHORT)rbits;
				shift = (USHORT)rshift;
				break;
			case 1:
				bits  = (USHORT)gbits;
				shift = (USHORT)gshift;
				break;
			case 2:
				bits  = (USHORT)bbits;
				shift = (USHORT)bshift;
				break;
			}

			for( j = 0; j < hdr.BytesPerLine; j ++ )
			{
				fc = fgetc( f );
				if( ( fc & 0xC0 ) == 0xC0)
				{
					l = (fc & 0x3F);

					fc = fgetc( f );

					col = ( fc & bits ) << shift;

					_asm
					{
						mov ecx, l
						InnerLoop:
						mov eax,[edi]
						or  eax, col
						mov [edi],eax
						add edi, 4
						inc j
						dec cx
						jnz InnerLoop
						dec j
					};
				}
				else
				{
					col = ( fc & bits ) << shift;

					_asm
					{
						mov eax,[edi]
						or  eax, col
						mov [edi],eax
						add edi, 4
					};
				}
			}
		}

	}

	fclose( f );

	return temp;
}

TGAHeader LoadTGAHeader( FILE *f )
{
	TGAHeader temp;
	char t[18];

	fread( &t, 18, 1, f );

	temp.IDLength             = t[0] & 0xFF;
	temp.ColorMap             = t[1] & 0xFF;
	temp.ImageType            = t[2] & 0xFF;
	temp.ColorSpec_FirstEntry = ( ( t[ 4] & 0xFF ) << 8 ) + ( t[ 3] & 0xFF );
	temp.ColorSpec_Length     = ( ( t[ 6] & 0xFF ) << 8 ) + ( t[ 5] & 0xFF );
	temp.ColorSpec_EntrySize  = t[7] & 0xFF;
	temp.ImageSpec_XOrig      = ( ( t[ 9] & 0xFF ) << 8 ) + ( t[ 8] & 0xFF );
	temp.ImageSpec_YOrig      = ( ( t[11] & 0xFF ) << 8 ) + ( t[10] & 0xFF );
	temp.ImageSpec_Width      = ( ( t[13] & 0xFF ) << 8 ) + ( t[12] & 0xFF );
	temp.ImageSpec_Height     = ( ( t[15] & 0xFF ) << 8 ) + ( t[14] & 0xFF );
	temp.ImageSpec_Depth      = t[16] & 0xFF;
	temp.ImageSpec_Desc       = t[17] & 0xFF;

	return temp;
}

pSurface DXSurface::LoadTGA(char *Filename, int mem)
{
	switch( m_bpp )
	{
	case 16:
		return LoadTGA16( Filename, mem );
		break;
	case 24:
		return LoadTGA24( Filename, mem );
		break;
	case 32:
		return LoadTGA32( Filename, mem );
		break;
	}

	return NULL;
}

pSurface DXSurface::LoadTGA16(char *Filename, int mem)
{
	FILE *f;
	TGAHeader hdr;
	char t[256];
	char fc;
	pSurface temp = NULL;
	int i,j,k,l,m;
	DDPIXELFORMAT ddPixFormat;
	USHORT rbits, gbits, bbits;
	USHORT rshift, gshift, bshift;
	DDSURFACEDESC2 ddSurf;
	HRESULT ddrval;
	USHORT col, a;
	char Alpha;

	f = fopen( Filename, "rb" );
	if( f == NULL )
	{
		Log( __FILE__, __LINE__, "DXSurface::LoadTGA16() - Could not open file" );
		return NULL;
	}

	hdr = LoadTGAHeader( f );

	Alpha = hdr.ImageSpec_Desc & 0x0F;

	if( Alpha )
		Log( __FILE__, __LINE__, "Loading Alpha enabled image" );

	temp = CreateSurface( hdr.ImageSpec_Width, hdr.ImageSpec_Height, mem, Alpha ? true : false );
	if( temp == NULL )
	{
		Log( __FILE__, __LINE__, "DXSurface::LoadTGA16() - Could not create surface" );
		fclose( f );
		return NULL;
	}

	strcpy( temp->Name, Filename );

	MLog1( __FILE__, __LINE__, "Added surface", Filename );

	DXSTRUCTINIT( ddPixFormat );

	temp->lpDDS->GetPixelFormat( &ddPixFormat );
	strcpy(temp->Name, Filename);

	DXSTRUCTINIT( ddSurf );

	ddrval = temp->lpDDS->Lock( NULL, &ddSurf, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL );
	if( ddrval != DD_OK )
	{
		sprintf( t, "DXSurface::LoadTGA16() - Failed to lock surface : %s", DDErrorString( ddrval ) );
		Log( __FILE__, __LINE__, t );
	}

	temp->lpDDS->Unlock( NULL );

	rbits = gbits = bbits = rshift = gshift = bshift = 0;

	if( ( ddPixFormat.dwGBitMask >> 5 ) == 0x3F )
	{
		rbits = 0xF8;
		gbits = 0xFC;
		bbits = 0xF8;
		rshift = 8;
		gshift = 3;
		bshift = 3;
	}
	else
		if( ( ddPixFormat.dwGBitMask >> 5 ) == 0x3F )
		{
			rbits = 0xF8;
			gbits = 0xF8;
			bbits = 0xF8;
			rshift = 7;
			gshift = 3;
			bshift = 3;
		}

	k = 0;

	for( i = hdr.ImageSpec_Height - 1; i >= 0; i -- )
//	for( i = 50; i > 0; i -- )
	{
		m = ( i * ddSurf.lPitch );

		_asm
		{
			mov edi, ddSurf.lpSurface
			mov eax, m
			add edi, eax
		};

		for( j = 0; j < hdr.ImageSpec_Width; j ++ )
		{
			fc = fgetc( f );
			if( ( fc & 0x80 ) == 0x80)
			{
				l = (fc & 0x7F);

				col  = ( ( fgetc( f ) & 0xFF ) & bbits ) >> bshift;
				col += ( ( fgetc( f ) & 0xFF ) & gbits ) << gshift;
				col += ( ( fgetc( f ) & 0xFF ) & rbits ) << rshift;
				if( Alpha != 0 )
					a = fgetc( f ) & 0xFF;

				_asm
				{
					mov ecx, l
					inc ecx
					mov ax, col
					InnerLoop:
					mov [edi],ax
					add edi, 2
					inc j
					dec cx
					jnz InnerLoop
					dec j
				};
			}
			else
			{
				l = (fc & 0x7F);

				for( k = l + 1; k > 0; k -- )
				{
					col  = ( ( fgetc( f ) & 0xFF ) & bbits ) >> bshift;
					col += ( ( fgetc( f ) & 0xFF ) & gbits ) << gshift;
					col += ( ( fgetc( f ) & 0xFF ) & rbits ) << rshift;
					if( Alpha != 0 )
						a = fgetc( f ) & 0xFF;

					_asm
					{
						mov ax, col
						mov [edi],ax
						add edi, 2
					};

					j++;
				}

				j--;
			}
		}

	}

	fclose( f );

	return temp;
}

pSurface DXSurface::LoadTGA24(char *Filename, int mem)
{
	FILE *f;
	TGAHeader hdr;
	char t[256];
	char fc;
	pSurface temp = NULL;
	int i,j,k,l,m;
	DDPIXELFORMAT ddPixFormat;
	USHORT rbits, gbits, bbits;
	USHORT rshift, gshift, bshift;
	DDSURFACEDESC2 ddSurf;
	HRESULT ddrval;
	char r, g, b, a;
	char Alpha;

	f = fopen( Filename, "rb" );
	if( f == NULL )
	{
		Log( __FILE__, __LINE__, "DXSurface::LoadTGA24() - Could not open file" );
		return NULL;
	}

	hdr = LoadTGAHeader( f );

	Alpha = hdr.ImageSpec_Desc & 0x0F;

	if( Alpha )
		Log( __FILE__, __LINE__, "Loading Alpha enabled image" );

	temp = CreateSurface( hdr.ImageSpec_Width, hdr.ImageSpec_Height, mem, Alpha ? true : false );
	if( temp == NULL )
	{
		Log( __FILE__, __LINE__, "DXSurface::LoadTGA24() - Could not create surface" );
		fclose( f );
		return NULL;
	}

	strcpy( temp->Name, Filename );

	MLog1( __FILE__, __LINE__, "Added surface", Filename );

	DXSTRUCTINIT( ddPixFormat );

	temp->lpDDS->GetPixelFormat( &ddPixFormat );

	DXSTRUCTINIT( ddSurf );

	ddrval = temp->lpDDS->Lock( NULL, &ddSurf, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL );
	if( ddrval != DD_OK )
	{
		sprintf( t, "DXSurface::LoadTGA24() - Failed to lock surface : %s", DDErrorString( ddrval ) );
		Log( __FILE__, __LINE__, t );
	}

	temp->lpDDS->Unlock( NULL );

	rbits = gbits = bbits = rshift = gshift = bshift = 0;

	rbits = 0xFF;
	gbits = 0xFF;
	bbits = 0xFF;
	rshift = 16;
	gshift =  8;
	bshift =  0;

	k = 0;

	for( i = hdr.ImageSpec_Height - 1; i >= 0; i -- )
//	for( i = 50; i > 0; i -- )
	{
		m = ( i * ddSurf.lPitch );

		_asm
		{
			mov edi, ddSurf.lpSurface
			mov eax, m
			add edi, eax
		};

		for( j = 0; j < hdr.ImageSpec_Width; j ++ )
		{
			fc = fgetc( f );
			if( ( fc & 0x80 ) == 0x80)
			{
				l = (fc & 0x7F);

				b = ( ( fgetc( f ) & 0xFF ) & bbits );
				g = ( ( fgetc( f ) & 0xFF ) & gbits );
				r = ( ( fgetc( f ) & 0xFF ) & rbits );
				if( Alpha != 0 )
					a = fgetc( f ) & 0xFF;

				_asm
				{
					mov ecx, l
					inc ecx
					InnerLoop:
					mov al, b
					mov [edi],al
					add edi, 1
					mov al, g
					mov [edi],al
					add edi, 1
					mov al, r
					mov [edi],al
					add edi, 1
					inc j
					dec cx
					jnz InnerLoop
					dec j
				};
			}
			else
			{
				l = (fc & 0x7F);

				for( k = l + 1; k > 0; k -- )
				{

					b = ( ( fgetc( f ) & 0xFF ) & bbits );
					g = ( ( fgetc( f ) & 0xFF ) & gbits );
					r = ( ( fgetc( f ) & 0xFF ) & rbits );
					if( Alpha != 0 )
						a = fgetc( f ) & 0xFF;

					_asm
					{
						mov al, b
						mov [edi],al
						add edi, 1
						mov al, g
						mov [edi],al
						add edi, 1
						mov al, r
						mov [edi],al
						add edi, 1
					};

					j++;
				}

				j--;
			}
		}

	}

	fclose( f );

	return temp;
}

pSurface DXSurface::LoadTGA32(char *Filename, int mem)
{
	FILE *f;
	TGAHeader hdr;
	char t[256];
	char fc;
	pSurface temp = NULL;
	int i,j,k,l,m;
	DDPIXELFORMAT ddPixFormat;
	USHORT rbits, gbits, bbits;
	USHORT rshift, gshift, bshift;
	DDSURFACEDESC2 ddSurf;
	HRESULT ddrval;
	DWORD col, a;
	char Alpha;

	f = fopen( Filename, "rb" );
	if( f == NULL )
	{
		Log( __FILE__, __LINE__, "DXSurface::LoadTGA24() - Could not open file" );
		return NULL;
	}

	hdr = LoadTGAHeader( f );

	Alpha = hdr.ImageSpec_Desc & 0x0F;

	if( Alpha )
		Log( __FILE__, __LINE__, "Loading Alpha enabled image" );

	temp = CreateSurface( hdr.ImageSpec_Width, hdr.ImageSpec_Height, mem, Alpha ? true : false );
	if( temp == NULL )
	{
		Log( __FILE__, __LINE__, "DXSurface::LoadTGA24() - Could not create surface" );
		fclose( f );
		return NULL;
	}

	strcpy( temp->Name, Filename );

	MLog1( __FILE__, __LINE__, "Added surface", Filename );

	DXSTRUCTINIT( ddPixFormat );

	temp->lpDDS->GetPixelFormat( &ddPixFormat );

	DXSTRUCTINIT( ddSurf );

	ddrval = temp->lpDDS->Lock( NULL, &ddSurf, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL );
	if( ddrval != DD_OK )
	{
		sprintf( t, "DXSurface::LoadTGA24() - Failed to lock surface : %s", DDErrorString( ddrval ) );
		Log( __FILE__, __LINE__, t );
	}

	temp->lpDDS->Unlock( NULL );

	rbits = gbits = bbits = rshift = gshift = bshift = 0;

	rbits = 0xFF;
	gbits = 0xFF;
	bbits = 0xFF;
	rshift = 16;
	gshift =  8;
	bshift =  0;

	k = 0;

	for( i = hdr.ImageSpec_Height - 1; i >= 0; i -- )
//	for( i = 50; i > 0; i -- )
	{
		m = ( i * ddSurf.lPitch );

		_asm
		{
			mov edi, ddSurf.lpSurface
			mov eax, m
			add edi, eax
		};

		for( j = 0; j < hdr.ImageSpec_Width; j ++ )
		{
			fc = fgetc( f );
			if( ( fc & 0x80 ) == 0x80)
			{
				l = (fc & 0x7F);

				col  = ( ( fgetc( f ) & 0xFF ) & bbits ) << bshift;
				col += ( ( fgetc( f ) & 0xFF ) & gbits ) << gshift;
				col += ( ( fgetc( f ) & 0xFF ) & rbits ) << rshift;
				if( Alpha != 0 )
					a = fgetc( f ) & 0xFF;

				_asm
				{
					mov ecx, l
					inc ecx
					mov eax, col
					InnerLoop:
					mov [edi],eax
					add edi, 4
					inc j
					dec cx
					jnz InnerLoop
					dec j
				};
			}
			else
			{
				l = (fc & 0x7F);

				for( k = l + 1; k > 0; k -- )
				{

					col  = ( ( fgetc( f ) & 0xFF ) & bbits ) << bshift;
					col += ( ( fgetc( f ) & 0xFF ) & gbits ) << gshift;
					col += ( ( fgetc( f ) & 0xFF ) & rbits ) << rshift;
					if( Alpha != 0 )
						a = fgetc( f ) & 0xFF;

					_asm
					{
						mov eax, col
						mov [edi],eax
						xor eax, eax
						add edi, 4
					};

					j++;
				}

				j--;
			}
		}

	}

	fclose( f );

	return temp;
}

