#include "game.h"

#define SEEDS 100
#define PERC  50

WORD GetNumberOfBits( DWORD dwMask )
{
	WORD wBits = 0;
	while( dwMask )
	{
		dwMask = dwMask & ( dwMask - 1 );  
		wBits++;
	}
	return wBits;
}

bool Game::Init( HWND hWnd, HINSTANCE hInst )
{
	DDPIXELFORMAT ddPixFormat;
	bool retval;
	DDSURFACEDESC2 ddsd;
//	char t[256];
//	HRESULT ddrval;
	
//	retval = m_DX.InitDirectDraw( hWnd );
	retval = m_DX.InitDirectDraw( hWnd, 640, 480, 16, true );
	
	if( retval == true )
	{
		Log( __FILE__, __LINE__, "Loading MOUSE.TGA" );
		m_Mouse = m_DX.LoadTGA( "GRFX\\MOUSE.TGA", VideoMemory );
		if( m_Mouse == NULL )
		{
			Log( __FILE__, __LINE__, "Loading MOUSE.TGA - SystemMemory" );
			m_Mouse = m_DX.LoadTGA( "GRFX\\MOUSE.TGA", SystemMemory );
			if( m_Mouse == NULL )
				return false;
		}

		Log( __FILE__, __LINE__, "Loading LGTILES.TGA" );
		m_LTiles = m_DX.LoadTGA( "GRFX\\LGTILES.TGA", VideoMemory );
		if( m_LTiles == NULL )
		{
			Log( __FILE__, __LINE__, "Loading LGTILES.TGA - SystemMemory" );
			m_LTiles = m_DX.LoadTGA( "GRFX\\LGTILES.TGA", SystemMemory );
			if( m_LTiles == NULL )
				return false;
		}

		Log( __FILE__, __LINE__, "Loading SMTILES.TGA" );
		m_STiles = m_DX.LoadTGA( "GRFX\\SMTILES.TGA", VideoMemory );
		if( m_STiles == NULL )
		{
			Log( __FILE__, __LINE__, "Loading SMTILES.TGA - SystemMemory" );
			m_STiles = m_DX.LoadTGA( "GRFX\\SMTILES.TGA", SystemMemory );
			if( m_STiles == NULL )
				return false;
		}

		Log( __FILE__, __LINE__, "Loading BIGFONT.TGA" );
		m_BigFont = m_DX.LoadTGA( "GRFX\\BIGFONT.TGA", VideoMemory );
		if( m_BigFont == NULL )
		{
			Log( __FILE__, __LINE__, "Loading BIGFONT.TGA - SystemMemory" );
			m_BigFont = m_DX.LoadTGA( "GRFX\\BIGFONT.TGA", SystemMemory );
			if( m_BigFont == NULL )
				return false;
		}

		Log( __FILE__, __LINE__, "Loading SMLFONT.TGA" );
		m_SmlFont = m_DX.LoadTGA( "GRFX\\SMLFONT.TGA", VideoMemory );
		if( m_SmlFont == NULL )
		{
			Log( __FILE__, __LINE__, "Loading SMLFONT.TGA - SystemMemory" );
			m_SmlFont = m_DX.LoadTGA( "GRFX\\SMLFONT.TGA", SystemMemory );
			if( m_SmlFont == NULL )
				return false;
		}

		Log( __FILE__, __LINE__, "Loading LOBJECTS.TGA" );
		m_LObjects = m_DX.LoadTGA( "GRFX\\LOBJECTS.TGA", VideoMemory );
		if( m_LObjects == NULL )
		{
			Log( __FILE__, __LINE__, "Loading LOBJECTS.TGA - SystemMemory" );
			m_LObjects = m_DX.LoadTGA( "GRFX\\LOBJECTS.TGA", SystemMemory );
			if( m_LObjects == NULL )
				return false;
		}

		Log( __FILE__, __LINE__, "Loading SOBJECTS.TGA" );
		m_SObjects = m_DX.LoadTGA( "GRFX\\SOBJECTS.TGA", VideoMemory );
		if( m_SObjects == NULL )
		{
			Log( __FILE__, __LINE__, "Loading SOBJECTS.TGA - SystemMemory" );
			m_SObjects = m_DX.LoadTGA( "GRFX\\SOBJECTS.TGA", SystemMemory );
			if( m_SObjects == NULL )
				return false;
		}

		Log( __FILE__, __LINE__, "Loading LMM.TGA" );
		m_LMM = m_DX.LoadTGA( "GRFX\\LMM.TGA", VideoMemory );
		if( m_LMM == NULL )
		{
			Log( __FILE__, __LINE__, "Loading LMM.TGA - SystemMemory" );
			m_LMM = m_DX.LoadTGA( "GRFX\\LMM.TGA", SystemMemory );
			if( m_LMM == NULL )
				return false;
		}

		Log( __FILE__, __LINE__, "Loading SMM.TGA" );
		m_SMM = m_DX.LoadTGA( "GRFX\\SMM.TGA", VideoMemory );
		if( m_SMM == NULL )
		{
			Log( __FILE__, __LINE__, "Loading SMM.TGA - SystemMemory" );
			m_SMM = m_DX.LoadTGA( "GRFX\\SMM.TGA", SystemMemory );
			if( m_SMM == NULL )
				return false;
		}

		Log( __FILE__, __LINE__, "Loading Mouse Maps" );
		if( LoadMouseMaps() == false )
		{
			Log( __FILE__, __LINE__, "Failed to load Mouse Maps" );
			return false;
		}

		m_DX.SetColorKey( m_BigFont, 0, 0 );
		m_DX.SetColorKey( m_SmlFont, 0, 0 );

		switch( m_DX.m_bpp )
		{
		case 16:
			DXSTRUCTINIT( ddPixFormat );

			m_LTiles->lpDDS->GetPixelFormat( &ddPixFormat );

			if( GetNumberOfBits( ddPixFormat.dwGBitMask ) == 6 )
			{
				m_DX.SetColorKey( m_LObjects, 0x0000, 0x0000 );
				m_DX.SetColorKey( m_SObjects, 0x0000, 0x0000 );
				m_DX.SetColorKey( m_LTiles,   0x0000, 0x0000 );
				m_DX.SetColorKey( m_STiles,   0x0000, 0x0000 );
				m_DX.SetColorKey( m_Mouse,    0xF81F, 0xF81F );
			}
			else
			{
				m_DX.SetColorKey( m_LObjects, 0x0000, 0x0000 );
				m_DX.SetColorKey( m_SObjects, 0x0000, 0x0000 );
				m_DX.SetColorKey( m_LTiles,   0x0000, 0x0000 );
				m_DX.SetColorKey( m_STiles,   0x0000, 0x0000 );
				m_DX.SetColorKey( m_Mouse,    0x7C1F, 0x7C1F );
			}
			break;
		case 24:
		case 32:
			m_DX.SetColorKey( m_LObjects, 0x00000000, 0x00000000 );
			m_DX.SetColorKey( m_SObjects, 0x00000000, 0x00000000 );
			m_DX.SetColorKey( m_LTiles,   0x00000000, 0x00000000 );
			m_DX.SetColorKey( m_STiles,   0x00000000, 0x00000000 );
			m_DX.SetColorKey( m_Mouse,    0x00FF00FF, 0x00FF00FF );
			break;
		}

		DXSTRUCTINIT( ddsd );

		Log( __FILE__, __LINE__, "Init DXInput" );
		retval = m_DX.Input.DI_Init( hInst, hWnd );
		if( retval == false)
			Log( __FILE__, __LINE__, "DXInput init fail" );

		LoadMap( "testmap.raw" );

		m_Size = 0;
		m_OffX = 0;
		m_OffY = 0;

		GameState = STATETITLESCREEN;
	}

	Log( __FILE__, __LINE__, "GameInit() - OK" );

	return retval;
}

void Game::GameLoop()
{
	m_DX.ClrScreen();

//	switch( GameState )
//	{
//	case STATETITLESCREEN:
//		RenderTitleScreen();
//		break;
//	case STATEENDGAME:
//		Render();
//		RenderGameOver();
//		break;
//	case STATEGAME:
		Render();
//		break;
//	case STATEHISCORE:
//		RenderHiScore();
//		break;
//	}

	CheckKeyInput();

	DrawMouse();

	m_DX.Flip();
}

void Game::DeInit()
{
	int x, y;
	pMapTile tle;

	m_DX.Input.DI_Term();

	m_DX.DeInitDirectDraw();

	for( y = 0; y < 10; y ++ )
	{
		for( x = 0; x < 10; x ++ )
		{
			if( Map[y][x].Next )
			{
				while( Map[y][x].Next )
				{
					tle = Map[y][x].Next;
					Map[y][x].Next = tle->Next;
					delete tle;
				}
			}
		}
	}
}

void Game::Render()
{
	char t[256];
	int x, y, xp, yp, ypo;
	int xl, xh, yl, yh;
	pMapTile tle;
	RECT rc;
	static int d = 0;

	if( d > 1000 )
	{
		d = 0;
//		m_OffX = ( m_OffX ++ ) % 20;
	}

	d += DeltaTime;

	// Debug code start
	if( m_Size == 0 )
	{
		rc.top    = 256;
		rc.left   = 0;
		rc.right  = 64;
		rc.bottom = 320;
	}
	else
	{
		rc.top    = 128;
		rc.left   = 0;
		rc.right  = 32;
		rc.bottom = 160;
	}
/*
	for( y = -(m_OffY % ( m_Size == 1 ? 32 : 64 ) ); y < m_DX.m_ScreenHeight; y += ( m_Size == 1 ? 32 : 64 ) )
	{
		for( x = -(m_OffX % ( m_Size == 1 ? 32 : 64 ) ); x < m_DX.m_ScreenWidth; x += ( m_Size == 1 ? 32 : 64 ) )
		{
			m_DX.Blit2Screen( ( m_Size == 1 ? m_STiles : m_LTiles ), rc, x, y );
		}
	}
*/
	// Debug code end

//	xl = (      ( m_DX.m_ScreenWidth - m_OffX - 288 ) + (                   0 - m_OffY - 144 ) * 2 ) / ( m_Size == 1 ? 32 : 64 );
//	yl = ( -1 * (                  0 - m_OffX - 288 ) + (                   0 - m_OffY - 144 ) * 2 ) / ( m_Size == 1 ? 32 : 64 );
//	xh = (      (                  0 - m_OffX - 288 ) + ( m_DX.m_ScreenHeight - m_OffY - 144 ) * 2 ) / ( m_Size == 1 ? 32 : 64 );
//	yh = ( -1 * ( m_DX.m_ScreenWidth - m_OffX - 288 ) + ( m_DX.m_ScreenHeight - m_OffY - 144 ) * 2 ) / ( m_Size == 1 ? 32 : 64 );

	MouseMap( 0, 0, xl, y );
	MouseMap( m_DX.m_ScreenWidth, 0, x, yl );
	MouseMap( 0, m_DX.m_ScreenHeight, x, yh );
	MouseMap( m_DX.m_ScreenWidth, m_DX.m_ScreenHeight, xh, y );

	xl -= 1;
	xh += 5;
	yl -= 1;
	yh += 5;

//	m_DX.Blit2Screen( m_Back, 0, 0 );

//	m_DX.Blit2Screen( m_Tiles, 20, 40 );

//	for( y = 0; y < 200; y ++ )
	for( y = yl; y <= yh; y ++ )
	{
//		for( x = 0; x < 200; x ++ )
		for( x = xl; x <= xh; x ++ )
		{
			if( ( x >= 0 ) && ( x <= 200 ) && ( y >= 0 ) && ( y <= 200 ) )
			{
				tle = &Map[y][x];

				xp = ( ( x - y ) << ( 5 - m_Size ) );

				xp -= m_OffX;

				if( ( xp >= -( m_Size == 1 ? 32 : 64 ) ) && (xp <= 640 ) )
				{
					ypo = ( ( x + y ) << ( 4 - m_Size ) );// - ( tle->Height << 3);

					while( tle )
					{
						yp = ypo - ( tle->Height << ( 3 - m_Size ) );

						yp -= m_OffY;

						if( ( yp >= -( m_Size == 1 ? 32 : 64 ) ) && ( yp <= 480 + ( 3 * ( m_Size == 1 ? 32 : 64 ) ) ) )
						{
			//				if( tle->Floor != 0 )
			//					m_DX.DrawIsoGroundTile( m_Tiles, tle->Floor, xp, yp - ( tle->Height << 5 ), true );
			//				if( tle->LWall != 0 )
			//					m_DX.DrawIsoLWallTile(  m_Tiles, tle->LWall, xp, yp - ( tle->Height << 5 ), true );
			//				if( tle->RWall != 0 )
			//					m_DX.DrawIsoRWallTile(  m_Tiles, tle->RWall, xp, yp - ( tle->Height << 5 ), true );

							if( tle->Floor >= 20 )
							{
								rc.top    = ( ( tle->Floor - 20 ) / 20 ) << ( 7 - m_Size );
								rc.left   = ( ( tle->Floor - 20 ) % 20 ) << ( 7 - m_Size );

								rc.right  = rc.left + ( m_Size == 1 ? 64 : 128 );
								rc.bottom = rc.top + ( m_Size == 1 ? 64 : 128 );


								m_DX.Blit2Screen( ( m_Size == 1 ? m_SObjects : m_LObjects ), rc, xp - ( m_Size == 1 ? 16 : 32 ), yp - ( m_Size == 1 ? 72 : 144 ) );
							}
							else
							{
								rc.top    = ( tle->Floor / 20 ) << ( 6 - m_Size );
								rc.left   = ( tle->Floor % 20 ) << ( 6 - m_Size );

								if( tle->Height > 0 )
									rc.top += ( m_Size == 1 ? 32 : 64 );

								rc.right  = rc.left + ( m_Size == 1 ? 32 : 64 );
								rc.bottom = rc.top + ( m_Size == 1 ? 32 : 64 );


								m_DX.Blit2Screen( ( m_Size == 1 ? m_STiles : m_LTiles ), rc, xp, yp - ( m_Size == 1 ? 32 : 64 ) );
							}
						}
						tle = tle->Next;
					}
				}
			}
		}
	}

	rc.top    = 0;
	rc.left   = 15 << 6;
	rc.right  = rc.left + 64;
	rc.bottom = rc.top + 64;

	if ( m_Size == 0 )
	{
		for( y = (-m_OffY) % 32; y <= m_DX.m_ScreenHeight; y += 32 )
		{
			for( x = (-m_OffX) % 64; x <= m_DX.m_ScreenWidth; x += 64 )
			{
				m_DX.Blit2Screen( m_LTiles, rc, x, y );
			}
		}
	}
	else
	{
		for( y = (-m_OffY) % 16; y <= m_DX.m_ScreenHeight; y += 16 )
		{
			for( x = (-m_OffX) % 32; x <= m_DX.m_ScreenWidth; x += 32 )
			{
				m_DX.Blit2Screen( m_LTiles, rc, x, y );
			}
		}
	}

	sprintf( t, "FPS(%d)", Fps );

	DrawString( 1, "Q to Quit", -1, 0 );
	DrawString( 2, t, 0, 0 );
}

void Game::DrawLetter(int x, int y, char Ltr)
{
	RECT rct;
	int tile = 0;

	if( ( Ltr >= '0' ) && ( Ltr <= '9' ) )
		tile = Ltr - 22;
	if( ( Ltr >= 'A' ) && ( Ltr <= 'Z' ) )
		tile = Ltr - 65;
	if( ( Ltr >= 'a' ) && ( Ltr <= 'z' ) )
		tile = Ltr - 97;
	if( Ltr == '.' )
		tile = 36;
	if( Ltr == '\'' )
		tile = 37;
	if( Ltr == '!' )
		tile = 38;
	if( ( Ltr == '(' ) || ( Ltr == '-' ) )
		tile = 39;
	if( Ltr == ')' )
		tile = 40;
	if( Ltr == '?' )
		tile = 41;

	rct.top    = ( tile / 10 ) << 3;
	rct.left   = ( tile % 10 ) << 3;
	rct.right  = rct.left + 8;
	rct.bottom = rct.top + 8;

	if( Ltr != ' ' )
		m_DX.Blit2Screen( m_SmlFont, rct, x, y );
}

void Game::DrawBigLetter(int x, int y, char Ltr)
{
	RECT rct;
	int tile = 0;

	if( ( Ltr >= '0' ) && ( Ltr <= '9' ) )
		tile = Ltr - 22;
	if( ( Ltr >= 'A' ) && ( Ltr <= 'Z' ) )
		tile = Ltr - 65;
	if( ( Ltr >= 'a' ) && ( Ltr <= 'z' ) )
		tile = Ltr - 97;
	if( Ltr == '.' )
		tile = 36;
	if( Ltr == '\'' )
		tile = 37;
	if( Ltr == '!' )
		tile = 38;
	if( Ltr == '(' )
		tile = 39;
	if( Ltr == ')' )
		tile = 40;
	if( Ltr == '?' )
		tile = 41;

	rct.top    = ( tile / 10 ) << 4;
	rct.left   = ( tile % 10 ) << 4;
	rct.right  = rct.left + 16;
	rct.bottom = rct.top + 16;

	if( Ltr != ' ' )
		m_DX.Blit2Screen( m_BigFont, rct, x, y );
}

void Game::CheckKeyInput( void )
{
	static int MapDelay = 0;
	static DWORD MoveTime = 0;

	CurTime   = timeGetTime();
	DeltaTime = CurTime - LastTime;
	LastTime  = CurTime;
	FpsTime  += DeltaTime;
	MoveTime  += DeltaTime;

	FramesRendered++;

	if( FpsTime > 1000 )
	{
		Fps = FramesRendered;
		FramesRendered  = 0;
		FpsTime = 0;
	}

	m_DX.Input.ProcessKBInput();

	switch( GameState )
	{
	case STATETITLESCREEN:
		if( m_DX.Input.KeyPressed( DIK_Q ) )
			SendMessage( ghWnd, WM_DESTROY, 0, 0 );

		if( m_DX.Input.KeyPressed( DIK_H ) )
			GameState = STATEHISCORE;

		if( m_DX.Input.KeyPressed( DIK_S ) )
			m_Size = 1;

		if( m_DX.Input.KeyPressed( DIK_D ) )
			m_Size = 0;

		if( m_DX.Input.KeyPressed( DIK_LEFT ) )
			m_OffX -= 4;

		if( m_DX.Input.KeyPressed( DIK_RIGHT ) )
			m_OffX += 4;

		if( m_DX.Input.KeyPressed( DIK_UP ) )
			m_OffY -= 2;

		if( m_DX.Input.KeyPressed( DIK_DOWN ) )
			m_OffY += 2;

		break;
	case STATEGAME:
		if( m_DX.Input.KeyPressed( DIK_ESCAPE ) )
			GameState = STATETITLESCREEN;
		break;
	case STATEHISCORE:
		if( m_DX.Input.KeyPressed( DIK_ESCAPE ) )
			GameState = STATETITLESCREEN;
	}
}

void Game::DrawString( int Size, char *Text, int x, int y )
{
	unsigned int i;

	if( ( Size == 1 ) && ( x == -1) )
	{
		x = 40 - ( strlen( Text ) >> 1 );
	}

	if( ( Size == 3 ) && ( x == -1) )
	{
		x = 20 - ( strlen( Text ) >> 1 );
	}

	for( i = 0; i < strlen( Text ); i ++ )
	{
		if( Size == 1 )
			DrawLetter( ( x + i ) << 3, y, Text[i] );
		if( Size == 2 )
			DrawLetter( x + ( i << 3 ), y, Text[i] );
		if( Size == 3 )
			DrawBigLetter( ( x + i ) << 4, y, Text[i] );
		if( Size == 4 )
			DrawBigLetter(  x + ( i << 4 ), y, Text[i] );
	}
}

void Game::RenderTitleScreen( void )
{
}

void Game::RenderGameOver( void )
{
}

void Game::RenderHiScore( void )
{
}

bool Game::LoadMap(char *Filename)
{
	FILE *fin;
	int x, y, h, t, t1, t2, t3, t4;

	x = rand() % (SIZEX + 1);
	y = rand() % (SIZEY + 1);
	t1 = x;
	t2 = y;

	if( ( fin = fopen( Filename, "rb" ) ) == NULL )
	{
		Log1( __FILE__, __LINE__, "Unable to load map", Filename );
		return false;
	}

	for( y = 0; y < SIZEY + 1; y ++ )
	{
		for( x = 0; x < SIZEX + 1; x ++ )
		{
			temp[y][x] = fgetc( fin );
		}
	}

	fclose( fin );

	for( y = 0; y < 250; y++ )
	{
		for( x = 0; x < 250; x++ )
		{
			h = temp[y][x];
			if( temp[y+1][x] < h )
				h = temp[y+1][x];
			if( temp[y][x+1] < h )
				h = temp[y][x+1];
			if( temp[y+1][x+1] < h )
				h = temp[y+1][x+1];

			t = 0;
			t1 = temp[y    ][x    ] - h;
			t2 = temp[y    ][x + 1] - h;
			t3 = temp[y + 1][x + 1] - h;
			t4 = temp[y + 1][x    ] - h;

			t = ( t1 == 0 ? 0 : 1 ) + ( t2 == 0 ? 0 : 2 ) + ( t3 == 0 ? 0 : 4 ) + ( t4 == 0 ? 0 : 8 );

			if( t1 == 2 )
				t = 16;
			if( t4 == 2 )
				t = 17;
			if( t3 == 2 )
				t = 18;
			if( t2 == 2 )
				t = 19;

//			Log3n( __FILE__, __LINE__, "Height is h at x,y:", h, x, y);
//			if( ( t1 > 1 ) || ( t1 < 0 ) ) Log3n( __FILE__, __LINE__, "t1 error", t1, x, y );
//			if( ( t2 > 1 ) || ( t2 < 0 ) ) Log3n( __FILE__, __LINE__, "t2 error", t2, x, y );
//			if( ( t3 > 1 ) || ( t3 < 0 ) ) Log3n( __FILE__, __LINE__, "t3 error", t3, x, y );
//			if( ( t4 > 1 ) || ( t4 < 0 ) ) Log3n( __FILE__, __LINE__, "t4 error", t4, x, y );

			Map[y][x].Floor = t;
			Map[y][x].Height = h;
			Map[y][x].Next = NULL;
		}
	}
	return true;
}

void Game::LoadTestMap()
{
	int c, x, y, h, t, t1, t2, t3, t4;

	srand( time( NULL ) );

	x = rand() % (SIZEX + 1);
	y = rand() % (SIZEY + 1);
	t1 = x;
	t2 = y;

	for( y = 0; y < SIZEY + 1; y ++ )
	{
		for( x = 0; x < SIZEX + 1; x ++ )
		{
			temp[y][x] = 0;
		}
	}

	for( h = 0; h < SEEDS; h ++ )
	{
		if( ( rand() % 4 ) == 0 )
		{
			x = rand() % (SIZEX + 1);
			y = rand() % (SIZEY + 1);
			ChangeHeight( x, y, 1 );
			t1 = x;
			t2 = y;
		}
		else
		{
			x = ( rand() % 51 ) - 25;
			y = ( rand() % 51 ) - 25;
			if( ( t1 + x ) > ( SIZEX + 1 ) )
				x = ( SIZEX + 1 ) - t1;
			if( ( t1 + x ) < 0 )
				x = -t1;
			if( ( t2 + y ) > ( SIZEY + 1 ) )
				y = ( SIZEY + 1 ) - t2;
			if( ( t2 + y ) < 0 )
				y = -t2;
			Line( t1, t2, t1 + x, t2 + y );
			t1 += x;
			t2 += y;
		}
	}

	for( c = 0; c < (SIZEX * SIZEY * PERC / 100); c++)
	{
		bool found = false;

		do
		{
			x = rand() % (SIZEX + 1);
			y = rand() % (SIZEY + 1);
			if( temp[y][x] != 0 ) continue;
			if( x > 0 && temp[y][x-1] > 0) found = true;
			if( x <= SIZEX && temp[y][x+1] > 0) found = true;
			if( y > 0 && temp[y-1][x] > 0) found = true;
			if( y <= SIZEY && temp[y+1][x] > 0) found = true;
		}
		while(!found);

		ChangeHeight( x, y, 1 );
	}

	for( y = 0; y < 250; y++ )
	{
		for( x = 0; x < 250; x++ )
		{
			h = temp[y][x];
			if( temp[y+1][x] < h )
				h = temp[y+1][x];
			if( temp[y][x+1] < h )
				h = temp[y][x+1];
			if( temp[y+1][x+1] < h )
				h = temp[y+1][x+1];

			t = 0;
			t1 = temp[y    ][x    ] - h;
			t2 = temp[y    ][x + 1] - h;
			t3 = temp[y + 1][x + 1] - h;
			t4 = temp[y + 1][x    ] - h;

			t = ( t1 == 0 ? 0 : 1 ) + ( t2 == 0 ? 0 : 2 ) + ( t3 == 0 ? 0 : 4 ) + ( t4 == 0 ? 0 : 8 );

			if( t1 == 2 )
				t = 16;
			if( t4 == 2 )
				t = 17;
			if( t3 == 2 )
				t = 18;
			if( t2 == 2 )
				t = 19;

//			Log3n( __FILE__, __LINE__, "Height is h at x,y:", h, x, y);
//			if( ( t1 > 1 ) || ( t1 < 0 ) ) Log3n( __FILE__, __LINE__, "t1 error", t1, x, y );
//			if( ( t2 > 1 ) || ( t2 < 0 ) ) Log3n( __FILE__, __LINE__, "t2 error", t2, x, y );
//			if( ( t3 > 1 ) || ( t3 < 0 ) ) Log3n( __FILE__, __LINE__, "t3 error", t3, x, y );
//			if( ( t4 > 1 ) || ( t4 < 0 ) ) Log3n( __FILE__, __LINE__, "t4 error", t4, x, y );

			Map[y][x].Floor = t;
			Map[y][x].Height = h;
			Map[y][x].Next = NULL;
			if( ( h > 0 ) && ( t == 0 ) )
			{
				Map[y][x].Next = new MapTile;
				Map[y][x].Next->Floor = 20;
				Map[y][x].Next->Height = 1;
				Map[y][x].Next->Next = NULL;
			}
		}
	}
}

void Game::ClearTestMap()
{
	int x, y;
	pMapTile tle;

	for( y = 0; y < 10; y ++ )
	{
		for( x = 0; x < 10; x ++ )
		{
			if( Map[y][x].Next )
			{
				while( Map[y][x].Next )
				{
					tle = Map[y][x].Next;
					Map[y][x].Next = tle->Next;
					delete tle;
				}
			}
		}
	}
}

void Game::DrawMouse()
{
	int x, y, xp, yp;
	RECT rc;

	MouseMap( giMouseX, giMouseY - ( m_Size == 1 ? 8 : 16 ), x, y );

	if( ( x >= 0 ) && ( x <= SIZEX ) && ( y >= 0 ) && ( y <= SIZEY ) )
	{
		rc.top    = 2 << ( 6 - m_Size );
		rc.left   = ( Map[y][x].Floor % 20 ) << ( 6 - m_Size );
		rc.right  = rc.left + ( m_Size == 1 ? 32 : 64 );
		rc.bottom = rc.top + ( m_Size == 1 ? 32 : 64 );

		xp = ( ( x - y ) << ( 5 - m_Size ) ) - m_OffX;

		yp = ( ( x + y ) << ( 4 - m_Size ) ) - ( Map[y][x].Height << ( 3 - m_Size ) ) - m_OffY;

		m_DX.Blit2Screen( ( m_Size == 1 ? m_STiles : m_LTiles ), rc, xp, yp - ( m_Size == 1 ? 32 : 64 ) );
	}

	rc.top    =  0;
	rc.left   =  0;
	rc.right  = 31;
	rc.bottom = 31;

	m_DX.Blit2Screen( m_Mouse, rc, giMouseX, giMouseY );
}

void Game::TileWalk(int &x, int &y, int Dir)
{
	switch( Dir )
	{
	case ISO_NORTH:
		{
			x--;
			y--;
			break;
		}
	case ISO_NORTHEAST:
		{
			y--;
			break;
		}
	case ISO_EAST:
		{
			x++;
			y--;
			break;
		}
	case ISO_SOUTHEAST:
		{
			x++;
			break;
		}
	case ISO_SOUTH:
		{
			x++;
			y++;
			break;
		}
	case ISO_SOUTHWEST:
		{
			y++;
			break;
		}
	case ISO_WEST:
		{
			x--;
			y++;
			break;
		}
	case ISO_NORTHWEST:
		{
			x--;
			break;
		}
	}
}

int Game::MM_LookUp( int x, int y, int t )
{
	int v;

	x %= ( m_Size == 1 ? 32 : 64 );
	y %= ( m_Size == 1 ? 32 : 64 );
	
	v = g_MouseMap[m_Size][t][y][x >> 1];

	if( ( x & 1 ) == 0 )
		v = v & 0x0F;
	else
		v = ( v >> 4 ) & 0x0F;

	return v;
}

void Game::MouseMap( int xp, int yp, int &mx, int &my )
{
	int m;
	int cx, cy;
	int fx, fy, tfy;
	int md, mc;

	// Adjustment for lower left corner offset.
	yp += ( m_Size == 1 ? 32 : 64 );

	// Step 1 : Screen -> World Co-Ords
	xp += m_OffX;
	yp += m_OffY;

	// Step 2 : Offset from world co-ords of tile (0,0)

	xp -= ( ( 0 - 0 ) << ( 5 - m_Size ) );
	yp -= ( ( 0 + 0 ) << ( 4 - m_Size ) );

	// Step 3 : Determine MouseMap Co-ords

	// Course co-ords
	cx = xp / ( m_Size == 1 ? 32 : 64 );
	cy = yp / ( m_Size == 1 ? 16 : 32 );

	// Fine co-ords
	fx = xp % ( m_Size == 1 ? 32 : 64 );
	fy = yp % ( m_Size == 1 ? 16 : 32 );

	// Adjust for my Unusual mouse maps...
	fy += ( m_Size == 1 ?  8 : 16 );

	// Adjust for negative co-ords
	if( fx < 0 )
	{
		fx += ( m_Size == 1 ) ? 32 : 64;
		cx --;
	}
	if( fy < 0 )
	{
		fy += ( m_Size == 1 ) ? 32 : 64;
		cy --;
	}

	// Step 4 : Perform Corse Tile walk
	mx = 0;
	my = 0;

	while( cy < 0 )
	{
		TileWalk( mx, my, ISO_NORTH );
		cy ++;
	}

	while( cy > 0 )
	{
		TileWalk( mx, my, ISO_SOUTH );
		cy --;
	}

	while( cx < 0 )
	{
		TileWalk( mx, my, ISO_WEST );
		cx ++;
	}

	while( cx > 0 )
	{
		TileWalk( mx, my, ISO_EAST );
		cx --;
	}

	// Step 5 : Use the lookup table for final tile walk...
	m = -1;
	mc = 4;

	while( ( m != MM_NONE ) && ( mc > 0 ) )
	{
		m = MM_LookUp( fx, 
			fy + ( Map[( my >= 0 ) ? my : 0 ][( mx >= 0 ) ? mx : 0 ].Height << ( 3 - m_Size ) ),
			Map[( my >= 0 ) ? my : 0 ][( mx >= 0 ) ? mx : 0 ].Floor );

		tfy = ( fy + ( Map[( my >= 0 ) ? my : 0 ][( mx >= 0 ) ? mx : 0 ].Height << ( 3 - m_Size ) ) );

		if( ( fx >= 0 ) && ( tfy >= 0 ) && ( fx < ( m_Size == 1 ? 32 : 64 ) ) && ( tfy < ( m_Size == 1 ? 32 : 64 ) ) )
		{
			switch( m )
			{
			case MM_NORTHWEST:
				{
					if( md == MM_SOUTHEAST )
						mc --;
					else
						mc = 4;
					TileWalk( mx, my, ISO_NORTHWEST );
					fx += ( m_Size == 1 ) ? 16 : 32; // Half x
					fy += ( m_Size == 1 ) ?  8 : 16; // Quarter y
					break;
				}
			case MM_NORTH:
				{
					if( md == MM_SOUTH )
						mc --;
					else
						mc = 4;
					TileWalk( mx, my, ISO_NORTH );
					fy += ( m_Size == 1 ) ? 16 : 32; // Half y
					break;
				}
			case MM_NORTHEAST:
				{
					if( md == MM_SOUTHWEST )
						mc --;
					else
						mc = 4;
					TileWalk( mx, my, ISO_NORTHEAST );
					fx -= ( m_Size == 1 ) ? 16 : 32; // Half x
					fy += ( m_Size == 1 ) ?  8 : 16; // Quarter y
					break;
				}
			case MM_SOUTHWEST:
				{
					if( md == MM_NORTHEAST )
						mc --;
					else
						mc = 4;
					TileWalk( mx, my, ISO_SOUTHWEST );
					fx += ( m_Size == 1 ) ? 16 : 32; // Half x
					fy -= ( m_Size == 1 ) ?  8 : 16; // Quarter y
					break;
				}
			case MM_SOUTH:
				{
					if( md == MM_NORTH )
						mc --;
					else
						mc = 4;
					TileWalk( mx, my, ISO_SOUTH );
					fy -= ( m_Size == 1 ) ? 16 : 32; // Half y
					break;
				}
			case MM_SOUTHEAST:
				{
					if( md == MM_NORTHWEST )
						mc --;
					else
						mc = 4;
					TileWalk( mx, my, ISO_SOUTHEAST );
					fx -= ( m_Size == 1 ) ? 16 : 32; // Half x
					fy -= ( m_Size == 1 ) ?  8 : 16; // Quarter y
					break;
				}
			}
		}
		else
		{
			if( tfy >= ( m_Size == 1 ? 32 : 64 ) )
			{
				m = MM_SOUTH;
				TileWalk( mx, my, ISO_SOUTH );
				fy -= ( m_Size == 1 ) ? 16 : 32;
			}
		}
		md = m;
	}
}


bool Game::LoadMouseMaps()
{
	FILE *fin;
	int x, y, o;
	char c1, c2, c;

/*
	0	White	6
	1	Blue	0
	2	Red		3
	3	Yellow	5
	4	Green	1
	5	Cyan	2
	6	Purple	4
*/

	fin = fopen( "Grfx\\smtiles.raw", "rb" );

	if( fin == NULL )
	{
		Log( __FILE__, __LINE__, "Unable to open smtiles.raw" );
		return false;
	}
	else
	{
		for( y = 0; y < 32; y++ )
		{
			for( o = 0; o < 20; o ++ )
			{
				for( x = 0; x < 16; x ++ )
				{
					c1 = fgetc( fin ) & 0x0F;
					c2 = fgetc( fin ) & 0x0F;

					switch ( c1 )
					{
					case 0:
						c = 0x10;
						break;
					case 1:
						c = 0x40;
						break;
					case 2:
						c = 0x50;
						break;
					case 3:
						c = 0x20;
						break;
					case 4:
						c = 0x60;
						break;
					case 5:
						c = 0x30;
						break;
					case 6:
						c = 0x00;
						break;
					}

					switch ( c2 )
					{
					case 0:
						c += 0x1;
						break;
					case 1:
						c += 0x4;
						break;
					case 2:
						c += 0x5;
						break;
					case 3:
						c += 0x2;
						break;
					case 4:
						c += 0x6;
						break;
					case 5:
						c += 0x3;
						break;
					case 6:
						c += 0x0;
						break;
					}

					g_MouseMap[1][o][y][x] = c;
				}
			}
		}

		fclose( fin );

		fin = fopen( "Grfx\\lgtiles.raw", "rb" );

		if( fin == NULL )
		{
			Log( __FILE__, __LINE__, "Unable to open lgtiles.raw" );
			return false;
		}
		else
		{
			for( y = 0; y < 64; y++ )
			{
				for( o = 0; o < 20; o ++ )
				{
					for( x = 0; x < 32; x ++ )
					{
						c1 = fgetc( fin ) & 0x0F;
						c2 = fgetc( fin ) & 0x0F;

						switch ( c1 )
						{
						case 0:
							c = 0x10;
							break;
						case 1:
							c = 0x40;
							break;
						case 2:
							c = 0x50;
							break;
						case 3:
							c = 0x20;
							break;
						case 4:
							c = 0x60;
							break;
						case 5:
							c = 0x30;
							break;
						case 6:
							c = 0x00;
							break;
						}

						switch ( c2 )
						{
						case 0:
							c += 0x1;
							break;
						case 1:
							c += 0x4;
							break;
						case 2:
							c += 0x5;
							break;
						case 3:
							c += 0x2;
							break;
						case 4:
							c += 0x6;
							break;
						case 5:
							c += 0x3;
							break;
						case 6:
							c += 0x0;
							break;
						}

						g_MouseMap[0][o][y][x] = c;
					}
				}
			}

			fclose( fin );
		}
	}

	return true;
}

void Game::Line(int sx, int sy, int ex, int ey)
{
	int dy = ey - sy;
	int dx = ex - sx;
	int stepx, stepy;
	int x = sx, y = sy;

    if (dy < 0)
	{
		dy = -dy;
		stepy = -1;
	}
	else
	{
		stepy = 1;
	}
    
	if (dx < 0)
	{
		dx = -dx;
		stepx = -1;
	}
	else
	{
		stepx = 1;
	}

    dy <<= 1;                                                  // dy is now 2*dy
    dx <<= 1;                                                  // dx is now 2*dx

//    raster.setPixel(pix, x0, y0);
//	ChangeHeight( x, y, 1 );
    if (dx > dy)
	{
        int fraction = dy - (dx >> 1);                         // same as 2*dy - dx
        while (x != ex)
		{
            if (fraction >= 0)
			{
                y += stepy;
                fraction -= dx;                                // same as fraction -= 2*dx
            }
            x += stepx;
            fraction += dy;                                    // same as fraction -= 2*dy
			ChangeHeight( x, y, 1 );
            //raster.setPixel(pix, x0, y0);
        }
    }
	else
	{
        int fraction = dx - (dy >> 1);
        while (y != ey)
		{
            if (fraction >= 0)
			{
                x += stepx;
                fraction -= dy;
            }
            y += stepy;
            fraction += dx;
			ChangeHeight( x, y, 1 );
            //raster.setPixel(pix, x0, y0);
        }
    }
}

void Game::ChangeHeight(int x, int y, int h)
{
	if( ( x >= 0 ) && ( x <= ( SIZEX + 1 ) ) && ( y >= 0 ) && ( y <= ( SIZEY + 1 ) ) )
	{
		if( h == 1 )
		{
			temp[y][x] ++;
			if( ( temp[y][x] - temp[y-1][x] ) > 1 )
				ChangeHeight( x, y-1, h );
			if( ( temp[y][x] - temp[y+1][x] ) > 1 )
				ChangeHeight( x, y+1, h );
			if( ( temp[y][x] - temp[y][x-1] ) > 1 )
				ChangeHeight( x-1, y, h );
			if( ( temp[y][x] - temp[y][x+1] ) > 1 )
				ChangeHeight( x+1, y, h );
		}
	}
}

