// -------------------------
// Windows 32 bit window init functions
// -------------------------

#include <windows.h>
#include <windowsx.h>
#include "resource.h"
#include "macros.h"
#include "game.h"

LRESULT CALLBACK WindowFunc( HWND, UINT, WPARAM, LPARAM );

char szWinName[] = "IsoEngineWindow";

HWND ghWnd;
unsigned int giMouseX, giMouseY, giMouseButtons;

bool bActive;

Game myGame;

bool InitWindow( HINSTANCE hInst )
{
	WNDCLASSEX wcl;

	wcl.cbSize = sizeof( WNDCLASSEX );
	wcl.hInstance = hInst;
	wcl.lpszClassName = szWinName;
	wcl.lpfnWndProc = WindowFunc;
	wcl.style = CS_HREDRAW | CS_VREDRAW;
	wcl.hIcon = LoadIcon( hInst, MAKEINTRESOURCE( IDI_APPICON ) );
	wcl.hIconSm = LoadIcon( hInst, MAKEINTRESOURCE( IDI_APPICON ) );
	wcl.hCursor = LoadCursor( hInst, MAKEINTRESOURCE( IDC_BLANK ) );
	wcl.lpszMenuName = NULL;
	wcl.cbClsExtra = 0;
	wcl.cbWndExtra = 0;
	wcl.hbrBackground = (HBRUSH) GetStockObject(BLACK_BRUSH);

	if( !RegisterClassEx( &wcl ) )
		return FALSE;

	ghWnd = CreateWindow( szWinName, "IsoEngine Window", WS_BORDER | WS_CAPTION | WS_SYSMENU | WS_VISIBLE, 
		0, 0, 640, 480, HWND_DESKTOP, NULL, hInst, NULL );

	if( ghWnd == NULL )
		return FALSE;

	ShowWindow( ghWnd, SW_SHOW );
	UpdateWindow( ghWnd );

	return TRUE;
}

int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE hPrevIntance, LPSTR lpszArgs, int nWinMode )
{
	bool bRetVal;
	MSG msg;

	InitLog();

	Log( __FILE__, __LINE__, "Init Window" );
	bRetVal = InitWindow( hInstance );

	Log( __FILE__, __LINE__, "Init Game Class" );
	bRetVal = myGame.Init( ghWnd, hInstance );

	if( bRetVal == TRUE )
	{
		Log( __FILE__, __LINE__, "Main Loop" );
		while(1)
		{
			if( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) )
			{
				if( msg.message == WM_QUIT )
					break;
				TranslateMessage( &msg );
				DispatchMessage( &msg );
			}
			else
			{
				if( bActive )
					myGame.GameLoop();
			}
		}
	}

	myGame.DeInit();
	Log( __FILE__, __LINE__, "Normal Exit..." );
	return 0;
}

LRESULT CALLBACK WindowFunc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
	switch( msg )
	{
	case WM_MOUSEMOVE:
		giMouseX = LOWORD(lParam);
		giMouseY = HIWORD(lParam);
		giMouseButtons = wParam;
		break;
	case WM_ACTIVATE:
		if( LOWORD( wParam ) == WA_INACTIVE )
			bActive = false;
		else
			bActive = true;
		break;
	case WM_DESTROY:
		myGame.DeInit();
		PostQuitMessage( 0 );
		break;
	default:
		return DefWindowProc( hWnd, msg, wParam, lParam );
	}

	return 0;
}